import 'dart:convert';
import 'dart:io';

import 'package:dio/dio.dart';
import 'package:flutter/foundation.dart';
import 'package:hive/hive.dart';
import 'constants.dart';
import 'logger.dart';
import 'model/base_list_response.dart';
import 'model/base_object_response.dart';
import 'network/either.dart';
import 'network/fetch_failure.dart';

abstract class BaseService {
  final logger = getLogger('NetWorkTask');

  @protected
  Future<Either<F, T>> safeCall<F, T>({
    required Future<T> Function() call,
    required F Function() onNetworkError,
    required F Function(Response? response)
        onResponseError, // ignore: always_specify_types
    required F Function(Exception e) onUnknownError,
  }) async {
    try {
      final T result = await call();
      return right(result);
    } on DioError catch (e) {
      // logger.e("BaseService.safeCall: ", e);
      switch (e.type) {
        case DioErrorType.connectTimeout:
          return left(onNetworkError());
        case DioErrorType.sendTimeout:
          return left(onNetworkError());
        case DioErrorType.receiveTimeout:
          return left(onNetworkError());
        case DioErrorType.response:
          return left(onResponseError(e.response));
        case DioErrorType.cancel:
          return left(onNetworkError());
        case DioErrorType.other:
          if (e.error is SocketException) {
            return left(onNetworkError());
          }
          return left(onUnknownError(e));
      }
    }
  }

  @protected
  Future<Either<FetchFailure, T>> safeFetch<T>(
      Future<T> Function() call) async {
    return safeCall(
      call: call,
      onNetworkError: () => const FetchFailure.networkError(),
      onResponseError: (Response? response) {
        try {
          if (response != null && response.statusCode != null) {
            int? statusCode = response.statusCode ?? 00;
            var jsonString = json.encode(response.data["errors"]);
            var msg = List<String>.from(json.decode(jsonString));
            var data = response.data["code_name"];
            if (statusCode == 401) {
              return FetchFailure.login(msg.toList(), data);
            } else if (statusCode >= 400 && statusCode < 500) {
              try {
                return FetchFailure.warningError(statusCode, msg, data);
              } catch (e) {
                msg = [response.statusMessage.toString()];
                data = response.data.toString();
                return FetchFailure.warningError(statusCode, msg, data);
              }
            } else if (statusCode >= 500 && statusCode < 600) {
              try {
                return FetchFailure.serverError(statusCode, msg, data);
              } catch (e) {
                msg = [response.statusMessage.toString()];
                data = response.data.toString();
                return FetchFailure.warningError(statusCode, msg, data);
              }
            }
          }
          return FetchFailure.unknownError(101, "Unknown", "Unknown");
        } catch (e) {
          var status = response?.statusCode ?? 101;
          String msg = response?.statusMessage ?? "";
          var data = response?.statusMessage ?? "";
          return FetchFailure.serverError(status, [msg], data);
        }
      },
      onUnknownError: (Exception e) =>
          FetchFailure.unknownError(101, "Unknown", "Unknown"),
    );
  }

  //=================== SAFE CALL FORM DATA LIST =====================//

  Future<Either<FetchFailure, List<T>?>> safeCallListStatus<T>(
      {required Either<FetchFailure, BaseListResponse<T>?> response,
      Function? callBack,
      bool isShowError = true}) async {
    try {
      if (response.get != null) {
        if (response.get?.data?.isNotEmpty == true) {
          if (callBack != null) {
            callBack();
          }
          return await response.map((data) => data?.data);
        } else {
          if (isShowError) {
            return await left(const FetchFailure.noData());
          } else {
            return await left(const FetchFailure.nothing());
          }
        }
      } else {
        return await safeResponseBaseListLeftOrCrash<T>(
            isShowError: isShowError, response: response);
      }
    } catch (e) {
      return await safeResponseError<List<T>?>(isShowError: isShowError, e: e);
    }
  }

  Future<void> safeCallListStatusTest<T>(
      {required Either<FetchFailure, BaseListResponse<T>?> response,
      required Function(FetchFailure) funError,
      required Function(List<T>?) funDataServer,
      String? keyCache,
      Function(List<T>?)? funDataCache,
      bool isShowError = true}) async {
    try {
      if (funDataCache != null) {
        await safeResponseDataCache<List<T>>(
            keyCache: keyCache,
            funCallBack: (data) {
              isShowError = false;
              funDataCache(data);
            });
      }
      if (response.get != null) {
        await safeResponseDataServer<List<T>>(
            dataServer: response.get?.data,
            funCallBack: (data) async {
              if (data?.isNotEmpty == true) {
                return funDataServer(data);
              } else {
                return funDataServer([]);
              }
            });
      } else {
        funError(await safeResponseErrorTest(
            typeError: TypeError.ERROR_SERVER,
            errorServer: response.leftOrCrash,
            isShowError: isShowError));
      }
    } catch (e) {
      final failure = returnFailure(e);
      funError(await safeResponseErrorTest(
          typeError: TypeError.ERROR_UNKNOWN,
          errorServer: FetchFailure.crashError(100, failure, "Error"),
          isShowError: isShowError));
    }
  }

  //=================== SAFE CALL FORM DATA OBJECT =====================//

  Future<Either<FetchFailure, T?>> safeCallObjectStatus<T>(
      {required Either<FetchFailure, BaseObjectResponse<T?>> response,
      Function? callBack,
      Function(T?)? saveCache,
      bool? isCacheData = false,
      bool isShowError = true}) async {
    try {
      if (response.get != null) {
        if (response.get?.data != null) {
          if (callBack != null) callBack();
          if (saveCache != null) saveCache(response.get?.data);
          return await response.map((data) => data.data);
        } else {
          if (isShowError) {
            return await left(const FetchFailure.noData());
          } else {
            return await left(const FetchFailure.nothing());
          }
        }
      } else {
        if (isCacheData == true) isShowError = false;
        return await safeResponseBaseObjectLeftOrCrash<T?>(
            isShowError: isShowError, response: response);
      }
    } catch (e) {
      return await safeResponseError<T?>(isShowError: isShowError, e: e);
    }
  }

  Future<Either<FetchFailure, T>> safeResponseError<T>(
      {bool? isShowError = true, Object? e}) {
    if (isShowError == true) {
      print(e.toString());
      final failure = returnFailure(e);
      return Future.value(left(FetchFailure.crashError(100, failure, "Error")));
    } else {
      return Future.value(left(const FetchFailure.nothing()));
    }
  }

  Future<Either<FetchFailure, List<T>?>> safeResponseBaseListLeftOrCrash<T>(
      {bool? isShowError = true,
      required Either<FetchFailure, BaseListResponse<T>?> response}) async {
    return isShowError == true
        ? Future.value(left(response.leftOrCrash))
        : Future.value(left(FetchFailure.nothing()));
  }

  Future<Either<FetchFailure, T>> safeResponseBaseObjectLeftOrCrash<T>(
      {bool? isShowError = true,
      required Either<FetchFailure, BaseObjectResponse<T>> response}) async {
    return isShowError == true
        ? Future.value(left(response.leftOrCrash))
        : Future.value(left(FetchFailure.nothing()));
  }

  Future<Either<FetchFailure, T?>> safeResponseLeftOrCrash<T>(
      {bool? isShowError = true,
      required Either<FetchFailure, T?> response}) async {
    return isShowError == true
        ? Future.value(left(response.leftOrCrash))
        : Future.value(left(FetchFailure.nothing()));
  }

  // Future<void> safeCallObjectStatus<T>(
  //     {required Either<FetchFailure, BaseObjectResponse<T?>> response,
  //     required Function(FetchFailure) funError,
  //     required Function(T?) funDataServer,
  //     String? keyCache,
  //     Function(T?)? funDataCache,
  //     bool isShowError = true}) async {
  //   try {
  //     if (funDataCache != null) {
  //       await safeResponseDataCache<T>(
  //           keyCache: keyCache,
  //           funCallBack: (data) {
  //             isShowError = false;
  //             funDataCache(data);
  //           });
  //     }
  //     if (response.get != null) {
  //       safeResponseDataServer<T>(
  //           dataServer: response.get?.data,
  //           funCallBack: (data) async {
  //             if (data != null) {
  //               return funDataServer(data);
  //             } else {
  //               return funError(await safeResponseError(
  //                   typeError: TypeError.ERROR_NO_DATA,
  //                   isShowError: isShowError));
  //             }
  //           });
  //     } else {
  //       funError(await safeResponseError(
  //           typeError: TypeError.ERROR_SERVER,
  //           errorServer: response.leftOrCrash,
  //           isShowError: isShowError));
  //     }
  //   } catch (e) {
  //     final failure = returnFailure(e);
  //     funError(await safeResponseError(
  //         typeError: TypeError.ERROR_UNKNOWN,
  //         errorServer: FetchFailure.crashError(100, failure, "Error"),
  //         isShowError: isShowError));
  //   }
}

//=================== SAFE CALL FORM DATA MESSAGE =====================//

Future<Either<FetchFailure, String?>> safeCallMessageStatus<T>(
    {required Either<FetchFailure, BaseObjectResponse<T>> response,
    bool isData = false,
    bool isShowError = true}) async {
  try {
    if (response.get != null) {
      if (isData) {
        return response.map((data) => data.data.toString());
      } else {
        return response.map((data) => data.message.toString());
      }
    } else {
      return left(await safeResponseErrorTest(
          typeError: TypeError.ERROR_SERVER,
          errorServer: response.leftOrCrash,
          isShowError: isShowError));
    }
  } catch (e) {
    print(e.toString());
    final failure = returnFailure(e);
    return left(await safeResponseErrorTest(
        typeError: TypeError.ERROR_UNKNOWN,
        errorServer: FetchFailure.crashError(100, failure, "Error"),
        isShowError: isShowError));
  }
}

//=================== SAFE CALL FORM DATA CACHE =====================//

Future<Either<String, T>> safeCallCacheStatus<T>(T? response) async {
  try {
    if (response != null) {
      return await right(response);
    } else {
      return await left("Error");
    }
  } catch (e) {
    final failure = returnFailure(e);
    return left(failure);
  }
}

Future<Either<FetchFailure, T>> safeCallLocalCache<T>(T? response) async {
  try {
    if (response != null) {
      return await right(response);
    } else {
      return await left(const FetchFailure.nothing());
    }
  } catch (e) {
    return await left(const FetchFailure.nothing());
  }
}

//=================== SAFE CALL HIVE BOX CACHE =====================//

Future<void> saveCacheData<T>({required String key, T? data}) async {
  final cacheBox = await Hive.openBox(key);
  cacheBox.put(key, data);
}

Future<void> safeResponseDataCache<T>(
    {required String? keyCache, required Function(T?) funCallBack}) async {
  if (keyCache?.isNotEmpty == true) {
    final cacheData = await responseCacheData<T>(key: keyCache!);
    if (cacheData != null) {
      funCallBack(cacheData);
    }
  }
}

//=================== SAFE RESPONSE LEFT OR CRASH =====================//

Future<T?> responseCacheData<T>({required String key}) async {
  T? dataCache;
  final cacheBox = await Hive.openBox(key);
  try {
    if (cacheBox.isNotEmpty == true) {
      if (cacheBox.values.single != null) {
        dataCache = cacheBox.get(key);
        return Future.value(dataCache);
      }
    }
  } catch (e) {
    cacheBox.put(key, null);
    return Future.value(dataCache);
  }
  cacheBox.put(key, null);
  return Future.value(dataCache);
}

Future<void> safeResponseDataServer<T>({
  required T? dataServer,
  required Function(T?) funCallBack,
  String? keyCache,
  bool? isDataCache = false,
}) async {
  if (keyCache?.isNotEmpty == true) {
    saveCacheData<T>(key: keyCache!, data: dataServer);
  }
  funCallBack(dataServer);
}
//=================== SAFE RESPONSE ERROR =====================//

Future<FetchFailure> safeResponseErrorTest(
    {required TypeError typeError,
    FetchFailure? errorServer,
    FetchFailure? errorUnknow,
    bool? isShowError = true}) async {
  switch (typeError) {
    case TypeError.ERROR_SERVER:
      if (isShowError == true) {
        return await errorServer ?? FetchFailure.noData();
      } else {
        return await const FetchFailure.nothing();
      }
    case TypeError.ERROR_UNKNOWN:
      if (isShowError == true) {
        return await errorUnknow ?? FetchFailure.noData();
      } else {
        return await const FetchFailure.nothing();
      }
    default:
      if (isShowError == true) {
        return await const FetchFailure.noData();
      } else {
        return await const FetchFailure.nothing();
      }
  }
}

enum TypeError {ERROR_SERVER, ERROR_UNKNOWN }
